from typing import Iterable, Optional, Literal

from crtl.components.span_linking_scorer import AbstractSpanSet, EntSet, SpanSet
from spacy.training import Example
from spacy.util import registry


def rank_link(
    candidate_concept_ids: Iterable[str], correct_concept_id: str
) -> Optional[int]:
    """
    Rank the candidate concept IDs based on the correct concept ID.

    Args:
        candidate_concept_ids (Iterable[str]): An iterable of candidate concept IDs.
        correct_concept_id (str): The correct concept ID.

    Returns:
        Optional[int]: The rank of the correct concept ID among the candidate concept IDs, starting from 1.
            If the correct concept ID is not found in the candidate concept IDs, returns None.
    """
    for i, candidate_concept_id in enumerate(candidate_concept_ids):
        if candidate_concept_id == correct_concept_id:
            return i + 1
    return None


def get_link_ranks(span_set: AbstractSpanSet) -> list[Optional[int]]:
    """
    Get the link ranks for a given span set.

    Args:
        span_set (AbstractSpanSet): An instance of AbstractSpanSet or its subclass.

    Returns:
        list[Optional[int]]: A list of ranks for each span with identical boundaries in the span set.
            If a the reference spans concept id is not included in the prediction mention candidates, None is appended.
    """
    link_ranks = []
    for prediction_spans, reference_spans in span_set:
        for prediction_span in prediction_spans:
            for reference_span in reference_spans:
                if (
                    prediction_span.start == reference_span.start
                    and prediction_span.end == reference_span.end
                ):

                    candidate_concept_ids = [
                        candidate_mention.concept_id
                        for candidate_mention in prediction_span._.candidate_mentions
                    ]
                    rank = rank_link(candidate_concept_ids, reference_span.kb_id_)
                    link_ranks.append(rank)
    return link_ranks


def score_link_rank_list(n: int, link_ranks: Optional[list[Optional[int]]]):
    """
    Score the link rank list by calculating the accuracy at a given rank.

    Args:
        n (int): The rank to calculate the accuracy at.
        link_ranks (list[Optional[int]]): A list of integers corresponding to either the rank of the correct concept in the list or None if it is not contained within the lsit.

    Returns:
        float: The accuracy at the given rank.
    """
    total_correct = 0
    if link_ranks:
        for rank in link_ranks:
            if rank:
                if rank <= n:
                    total_correct += 1
        return total_correct / len(link_ranks)
    else:
        return None


def score_accuracy_at_n(
    examples: Iterable[Example],
    ns: list[int],
    ent_or_span: Literal["ent", "span"],
    span_key: Optional[str] = None,
):
    """
    Compute the rank-at-n score for a given set of examples.

    Args:
        examples (Iterable[Example]): An iterable of training examples.
        ns (list[int]): A list of ranks to calculate the accuracy at.
        ent_or_span (Literal["ent", "span"]): Specifies whether to use entity-based or span-based evaluation.
            Must be either "ent" or "span".
        span_key (str): The span key to use for span-based evaluation.

    Returns:
        dict: A dictionary containing the rank-at-n scores for each rank in ns.
            The keys are in the format "rank_at_<n>_score".
    """
    if ent_or_span == "ent":
        span_set: AbstractSpanSet = EntSet(examples)
    elif ent_or_span == "span":
        span_set = SpanSet(examples, span_key)
    else:
        raise ValueError("ent_or_span must be either 'ent' or 'span'")
    link_ranks = get_link_ranks(span_set)
    scores = {"rank_at_n_scores": {str(n): score_link_rank_list(n, link_ranks) for n in ns}}
    return scores


@registry.scorers("crtl.score_accuracy_at_n")
def get_accuracy_at_n_scorer(
    ns: list[int],
    ent_or_span: Literal["ent", "span"],
    span_key: Optional[str] = None,
):
    def score(examples, **kwargs):
        return score_accuracy_at_n(examples, ns, ent_or_span, span_key)

    return score
