from collections import defaultdict
import datetime
import json
import os
import pathlib
import tarfile
from typing import Dict, Iterable, List, NamedTuple, Tuple, Union
from spacy.util import registry
import numpy
import tempfile

import nmslib
import scipy
import scipy.sparse
from sklearn.feature_extraction.text import TfidfVectorizer
from crtl.lkb.lkb import LexicalKnowledgeBase
from crtl.s3_file_utils.s3_file_utils import cached_path


class MentionCandidate(NamedTuple):
    """
    A data class representing a candidate entity that a mention may be linked to.
    Parameters
    ----------
    concept_id : str, required.
        The canonical concept id in the KB.
    terms : List[str], required.
        The aliases that caused this entity to be linked.
    similarities : List[float], required.
        The cosine similarities from the mention text to the alias in tf-idf space.
    """

    concept_id: str
    similarity: float


class ANNCandidateGenerator:
    """
    Candidate generator that uses Approximate Nearest Neighbor (ANN) search to find mention candidates.

    This class takes a Lexical Knowledge Base (LKB), builds an ANN index of term vectors,
    and provides methods to generate mention candidates given input terms.

    Parameters:
        lkb (LexicalKnowledgeBase): The Lexical Knowledge Base containing term-to-concept mappings.
        term_index (list[str]): A list of terms corresponding to the vectors in the ANN index.
        ann_index: The ANN index used for nearest neighbor search.
        term_vectors: The term vectors used for similarity comparison.
        vectorizer (TfidfVectorizer): The TF-IDF vectorizer used to convert terms into vectors.

    """

    def __init__(
        self,
        lkb: LexicalKnowledgeBase,
        term_index: list[str],
        ann_index,  #: FloatIndex,
        term_vectors: scipy.sparse.spmatrix,
        vectorizer: TfidfVectorizer,
    ):
        self.lkb = lkb
        self._term_index = term_index
        self._ann_index = ann_index
        self._vectorizer = vectorizer
        self._vectors = term_vectors

    @classmethod
    def _create_vectorizer(
        cls, lkb: LexicalKnowledgeBase, min_df: int
    ) -> TfidfVectorizer:
        """
        Create a TF-IDF vectorizer for generating term vectors.

        Parameters:
        - lkb (LexicalKnowledgeBase): The LexicalKnowledgeBase instance containing term-to-concept mappings.
        - min_df (int): The minimum document frequency threshold for terms to be included in the vectorizer.

        Returns:
        - tfidf_vectorizer (TfidfVectorizer): The initialized TF-IDF vectorizer.

        Note:
        - The TF-IDF vectorizer is configured with the following settings:
            - Analyzer: "char_wb"
            - N-gram Range: (3, 3)
            - Minimum Document Frequency (min_df): The threshold for terms to be included in the vectorizer.
            - Data Type: numpy.float64

        """
        tfidf_vectorizer = TfidfVectorizer(
            analyzer="char_wb", ngram_range=(3, 3), min_df=min_df, dtype=numpy.float64
        )
        tfidf_vectorizer.fit_transform(lkb.term_to_cuis.keys())
        return tfidf_vectorizer

    @classmethod
    def initialize_candidate_generator(
        cls, lkb: LexicalKnowledgeBase, min_df: int = 10
    ):
        """
        Initializes the ANNCandidateGenerator by creating vectors, building an approximate nearest neighbor index,
        and returning an instance of the ANNCandidateGenerator.

        Parameters:
            cls (type): The class object of the ANNCandidateGenerator.
            lkb (LexicalKnowledgeBase): The LexicalKnowledgeBase object containing the term-to-concept mappings.
            min_df (int, optional): The minimum document frequency threshold for the TfidfVectorizer. Defaults to 10.

        Returns:
            ANNCandidateGenerator: An instance of the ANNCandidateGenerator initialized with the provided parameters.
        """
        # sort to allow indexing later uses
        terms = sorted(list(lkb.term_to_cuis.keys()))
        start_time = datetime.datetime.now()
        print(f"Creating  vectors...")
        vectorizer = cls._create_vectorizer(lkb, min_df)
        vectors = vectorizer.transform(terms)
        print(f"Created vectors.")
        print("Finding empty (all zeros) tfidf vectors")
        vectors, terms = cls._delete_empty_vectors(vectors, terms)
        ann_index = cls._create_ann_index(vectors)
        end_time = datetime.datetime.now()
        elapsed_time = end_time - start_time
        print(
            f"Creating candidate generator took {elapsed_time.total_seconds()} seconds"
        )
        return cls(lkb, terms, ann_index, vectors, vectorizer)

    @classmethod
    def _create_ann_index(cls, vectors: scipy.sparse.csr_matrix):
        """
        Creates an approximate nearest neighbor (ANN) index using NMSLIB.

        Parameters:
        - vectors (scipy.sparse.csr_matrix,): The term vectors used for similarity comparison.

        Returns:
        - ann_index ?FloatIndex: The initialized ANN index.

        Note:
        - This method uses the NMSLIB library to create the ANN index.
        - The index is constructed using the HNSW (Hierarchical Navigable Small World Graph) representation.
        - Hyperparameters guide: https://github.com/nmslib/nmslib/blob/ade4bcdc9dd3719990de2503871450b8a62df4a5/manual/methods.md
            - `M` (m_parameter): Set to the maximum recommended value. Improves recall at the expense of longer indexing time.
                We use the HNSW representation, which is constructed by consecutive insertion of elements in a random order
                by connecting them to M closest neighbors from the previously inserted elements.
                These connections serve as bridges between network hubs, improving overall graph connectivity.
                Larger values of M result in higher recall but slower index creation.
                For more details, refer to: https://arxiv.org/pdf/1603.09320.pdf
            - `C` (construction): Set to the maximum recommended value. Improves recall at the expense of longer indexing time.
                This parameter controls the number of data points that can be visited during the index construction phase.
                Increasing C can improve recall, but it also increases the time required for index creation.
        """
        m_parameter = 100
        # `C` for Construction. Set to the maximum recommended value
        # Improves recall at the expense of longer indexing time
        construction = 2000
        num_threads = 60  # set based on the machine
        index_params = {
            "M": m_parameter,
            "indexThreadQty": num_threads,
            "efConstruction": construction,
            "post": 0,
        }

        ann_index = nmslib.init(
            method="hnsw",
            space="cosinesimil_sparse",
            data_type=nmslib.DataType.SPARSE_VECTOR,
        )
        ann_index.addDataPointBatch(vectors)
        ann_index.createIndex(index_params, print_progress=True)
        return ann_index

    @classmethod
    def _delete_empty_vectors(
        cls, vectors: scipy.sparse.csr_matrix, terms: list[str]
    ) -> Tuple[scipy.sparse.csr_matrix, list[str]]:
        """
        Remove empty TF-IDF vectors from the input vectors and corresponding terms.

        Parameters:
            vectors (scipy.sparse.csr_matrix): Input TF-IDF vectors.
            terms (list[str]): Corresponding terms.

        Returns:
            vectors, terms (Tuple[scipy.sparse.csr_matrix, list[str]]): Tuple containing the updated vectors and terms.

        Note:
        - Empty TF-IDF vectors are identified by summing the values along each row and checking if the sum is zero.
        - The function removes the empty vectors and updates the terms accordingly.
        - The updated vectors and terms are returned as a tuple.

        """
        empty_tfidfs_boolean_flags = numpy.array(vectors.sum(axis=1) != 0).reshape(-1)
        number_of_non_empty_tfidfs = sum(
            empty_tfidfs_boolean_flags == False
        )  # noqa: E712
        total_number_of_tfidfs = numpy.size(vectors, 0)
        print(
            f"Deleting {number_of_non_empty_tfidfs}/{total_number_of_tfidfs} aliases because their tfidf is empty"
        )
        # remove empty tfidf vectors, otherwise nmslib will crash
        terms = [
            alias for alias, flag in zip(terms, empty_tfidfs_boolean_flags) if flag
        ]
        vectors = vectors[empty_tfidfs_boolean_flags]
        assert len(terms) == numpy.size(vectors, 0)
        return vectors, terms

    def nmslib_knn_with_zero_vectors(
        self, vectors: numpy.ndarray, k: int
    ) -> Tuple[numpy.ndarray, numpy.ndarray]:
        """
        Perform k-nearest neighbor search using an ANN index, while handling empty vectors.

        This function serves as a wrapper around `ann_index.knnQueryBatch` to address the issue of empty vectors.
        It works as follows:
        - Remove empty vectors from the `vectors` array.
        - Call `ann_index.knnQueryBatch` with the non-empty vectors, which returns `neighbors` and `distances`.
        - Extend the `neighbors` and `distances` arrays with `None` values in place of the empty vectors.
        - Return the extended arrays of neighbors and distances.

        Parameters:
        - vectors (numpy.ndarray): The input array of vectors for k-nearest neighbor search.
        - k (int): The number of nearest neighbors to retrieve.

        Returns:
        - Tuple[numpy.ndarray, numpy.ndarray]: A tuple containing the extended arrays of neighbors and distances.

        Note:
        - The input `vectors` array should be a numpy.ndarray.
        - The `k` parameter specifies the number of nearest neighbors to retrieve.
        """
        empty_vectors_boolean_flags = numpy.array(vectors.sum(axis=1) != 0).reshape(-1)
        empty_vectors_count = vectors.shape[0] - sum(empty_vectors_boolean_flags)

        # init extended_neighbors with a list of Nones
        extended_neighbors = numpy.empty(
            (len(empty_vectors_boolean_flags),), dtype=object
        )
        extended_distances = numpy.empty(
            (len(empty_vectors_boolean_flags),), dtype=object
        )

        if vectors.shape[0] - empty_vectors_count == 0:
            return extended_neighbors, extended_distances

        # remove empty vectors before calling `ann_index.knnQueryBatch`
        vectors = vectors[empty_vectors_boolean_flags]

        # call `knnQueryBatch` to get neighbors
        original_neighbours = self._ann_index.knnQueryBatch(vectors, k=k)

        neighbors, distances = zip(
            *[(x[0].tolist(), x[1].tolist()) for x in original_neighbours]
        )
        neighbors = list(neighbors)
        distances = list(distances)

        # neighbors need to be converted to an np.array of objects instead of ndarray of dimensions len(vectors)xk
        # Solution: add a row to `neighbors` with any length other than k. This way, calling np.array(neighbors)
        # returns an np.array of objects
        neighbors.append([])
        distances.append([])
        # interleave `neighbors` and Nones in `extended_neighbors`
        extended_neighbors[empty_vectors_boolean_flags] = numpy.array(
            neighbors, dtype=object
        )[:-1]
        extended_distances[empty_vectors_boolean_flags] = numpy.array(
            distances, dtype=object
        )[:-1]

        return extended_neighbors, extended_distances

    def to_disk(self, path: Union[pathlib.Path, str]):
        """
        Save the state of the object to disk. This include:
        - The vector representations of terms as a SciPy sparse matrix (vectors.npz).
        - The lookup object (lkb).
        - The ANN index for efficient nearest neighbors search (ann_index).
        - A list of terms corresponding to the vectors (term_index).
        - A JSON file with specific parameters of the vectorizer (vectorizer_params.json).

        Parameters:
        - path (Union[pathlib.Path, str]): The location where the object's state should be saved. It can be a pathlib.Path object or a string representing a path.

        Returns:
        - None

        Note:
        - The method creates the directory path if it does not exist, using the `pathlib.Path.mkdir` method with `exist_ok=True`.
        - The method uses `scipy.sparse.save_npz` to save the vectors, nmslib's `saveIndex` method to save the ANN index, and standard file I/O operations for the remaining elements.
        """
        path = pathlib.Path(path)
        path.mkdir(exist_ok=True)
        scipy.sparse.save_npz(path / "vectors.npz", self._vectors.astype(numpy.float32))
        self.lkb.to_disk(path / "lkb")
        self._ann_index.saveIndex(str(path / "ann_index"))
        with open(path / "term_index", "w") as f:
            for term in self._term_index:
                f.write(term + "\n")
        with open(path / "vectorizer_params.json", "w") as f:
            json.dump({"min_df": self._vectorizer.get_params()["min_df"]}, f)

    @classmethod
    def from_disk(cls, path: pathlib.Path):
        """
        Load a previously saved state of the object from disk. It reads and loads the following elements:
        - The vector representations of terms as a SciPy sparse matrix (vectors.npz).
        - lkb: LexicalKnowledgeBase.
        - The ANN index for efficient nearest neighbors search (ann_index).
        - A list of terms corresponding to the vectors (term_index).
        - A JSON file with specific parameters of the vectorizer (vectorizer_params.json).

        Parameters:
        - path (pathlib.Path): The location from where the object's state should be loaded.

        Returns:
        - An instance of the class with its state restored from the disk.

        Note:
        - The method uses `scipy.sparse.load_npz` to load the vectors, nmslib's `loadIndex` method to load the ANN index, and standard file I/O operations for the remaining elements.
        """
        term_vectors = scipy.sparse.load_npz(path / "vectors.npz").astype(numpy.float32)
        ann_index = nmslib.init(
            method="hnsw",
            space="cosinesimil_sparse",
            data_type=nmslib.DataType.SPARSE_VECTOR,
        )
        ann_index.addDataPointBatch(term_vectors)
        ann_index.loadIndex(str(path / "ann_index"))
        lkb = LexicalKnowledgeBase.from_disk(path / "lkb")
        term_index = [term.strip() for term in open(path / "term_index").readlines()]
        min_df = json.load(open(path / "vectorizer_params.json"))["min_df"]
        vectorizer = cls._create_vectorizer(lkb, min_df)
        return cls(lkb, term_index, ann_index, term_vectors, vectorizer)

    def set_runtime_parameters(self, runtime_params: Dict):
        """
        Adjust the runtime parameters of the Approximate Nearest Neighbors (ANN) index.

        This function sets the query time parameters of the ANN index, specifically the `efSearch` parameter
        which controls the trade-off between search speed and recall in the HNSW algorithm used by the ANN index.

        Parameters:
        - runtime_params (Dict): A dictionary containing the runtime parameters to set. This should include a key "ef_search".

        Note:
        - The `efSearch` parameter should be included in `runtime_params` dictionary. It sets the balance between
        accuracy and speed of nearest neighbor searches. Higher values result in more accurate results but slower searches.
        """
        ef_search = runtime_params["ef_search"]
        if ef_search:
            self._ann_index.setQueryTimeParams({"efSearch": ef_search})

    def __call__(
        self, terms: Iterable[str], k: int
    ) -> Iterable[Iterable[MentionCandidate]]:
        """
        Performs a k-nearest neighbor search for a list of terms.

        The function first transforms the terms into vectors using the stored vectorizer.
        Then, it retrieves the nearest neighbors for each vector. If the list of terms is empty,
        it returns an empty list.

        Parameters:
        - terms (Iterable[str]): A list of terms for which to find nearest neighbors.
        - k (int): The number of nearest neighbors to retrieve.

        Returns:
        - Iterable[Iterable[MentionCandidate]]: A list of lists of MentionCandidate objects, representing
            the k nearest neighbors for each term.
        """
        # tfidf vectorizer crashes on an empty array, so we return early here
        terms = list(terms)
        if len(terms) == 0:
            return []
        vectors = self._vectorizer.transform(terms)
        # `ann_index.knnQueryBatch` crashes if one of the vectors is all zeros.
        # `nmslib_knn_with_zero_vectors` is a wrapper around `ann_index.knnQueryBatch` that addresses this issue.
        batch_mention_candidates = self._get_batch_mention_candidates(vectors, k)
        return batch_mention_candidates

    def _get_batch_mention_candidates(
        self, vectors: numpy.ndarray, k: int
    ) -> List[Iterable[MentionCandidate]]:
        """
        Retrieves mention candidates for a batch of input vectors.

        Parameters:
            vectors (numpy.ndarray): The input vectors for which mention candidates are retrieved.
            k (int): The number of nearest neighbors to consider for each vector.

        Returns:
            List[Iterable[MentionCandidate]]: A list of iterables, where each iterable contains `MentionCandidate` objects
                representing candidate entities for a given batch of input vectors.
        """
        batch_neighbors, batch_distances = self.nmslib_knn_with_zero_vectors(vectors, k)
        batch_mention_candidates = []
        for neighbors, distances in zip(batch_neighbors, batch_distances):
            mention_candidates = self._get_mention_candidates(neighbors, distances)
            batch_mention_candidates.append(mention_candidates)
        return batch_mention_candidates

    def _get_neighbor_terms(
        self, terms: Iterable[str], k: int
    ) -> Iterable[Tuple[str, float]]:
        """
        Perform a k-nearest neighbor search for a list of terms and returns tuples of the term and its corresponding distance.

        The function first transforms the terms into vectors using the stored vectorizer.
        Then, it retrieves the nearest neighbors for each vector. If the list of terms is empty,
        it returns an empty list.

        Parameters:
        - terms (Iterable[str]): A list of terms for which to find nearest neighbors.
        - k (int): The number of nearest neighbors to retrieve.

        Returns:
        - Iterable[Tuple[str, float]]: A list of tuples where each tuple contains a term and its corresponding distance.
        """
        # tfidf vectorizer crashes on an empty array, so we return early here
        terms = list(terms)
        if len(terms) == 0:
            return []
        vectors = self._vectorizer.transform(terms)

        # Call the nmslib_knn_with_zero_vectors method to retrieve nearest neighbors and distances
        # `ann_index.knnQueryBatch` crashes if one of the vectors is all zeros.
        # `nmslib_knn_with_zero_vectors` is a wrapper around `ann_index.knnQueryBatch` that addresses this issue.
        batch_neighbors, batch_distances = self.nmslib_knn_with_zero_vectors(vectors, k)
        results = []
        for neighbors, distances in zip(batch_neighbors, batch_distances):
            if neighbors is None:
                neighbors = []
            if distances is None:
                distances = []
            results.append(
                sorted(
                    [
                        (self._term_index[neighbor_index], distance)
                        for neighbor_index, distance in zip(neighbors, distances)
                    ],
                    key=lambda x: x[1],
                )
            )
        return results

    def _get_mention_candidates(self, neighbors, distances):
        """
        Create a list of mention candidates given a list of neighbors and their corresponding distances.

        Parameters:
        - neighbors (List[int]): A list of neighbor indices.
        - distances (List[float]): A list of distances corresponding to the neighbors.

        Returns:
        - List[MentionCandidate]: A sorted list of MentionCandidate objects, representing candidate entities for given neighbors.
        """
        if neighbors is None:
            neighbors = []
        if distances is None:
            distances = []
        concept_to_mentions: Dict[str, List[str]] = defaultdict(list)
        concept_to_similarities: Dict[str, List[float]] = defaultdict(list)
        for neighbor_index, distance in zip(neighbors, distances):
            mention = self._term_index[neighbor_index]
            concepts_for_mention = self.lkb.term_to_cuis[mention]
            for concept_id in concepts_for_mention:
                concept_to_mentions[concept_id].append(mention)
                concept_to_similarities[concept_id].append(1.0 - distance)
        mention_candidates = [
            MentionCandidate(concept, max(concept_to_similarities[concept]))
            for concept in concept_to_mentions.keys()
        ]
        mention_candidates = sorted(
            mention_candidates, key=lambda c: c.similarity, reverse=True
        )
        return mention_candidates

    @classmethod
    def from_gzip(cls, path: pathlib.Path):
        """
        Load an instance of the class from a gzipped file.

        Parameters:
        - path (pathlib.Path): The path to the gzipped file.

        Returns:
        - ANNCandidateGenerator: An instance of the class.
        """
        with tempfile.TemporaryDirectory() as tmpdirname:
            tar = tarfile.open(path)
            file_name = tar.getnames()[0]
            tar.extractall(tmpdirname)
            return cls.from_disk(pathlib.Path(tmpdirname) / file_name)


@registry.misc("load_ann_candidate_generator")
def load_ann_candidate_generator(url_str_or_path: Union[str, pathlib.Path]):
    """
    Load an ANNCandidateGenerator from a file or URL.

    If the file has a ".gz" extension, it is assumed to be a gzipped file, and the `from_gzip` method is used.
    Otherwise, the `from_disk` method is used.

    Parameters:
    - url_str_or_path (Union[str, pathlib.Path]): The path or URL to the file.

    Returns:
    - ANNCandidateGenerator: An instance of ANNCandidateGenerator loaded from the specified file.
    """
    path: pathlib.Path = cached_path(url_str_or_path)
    if path.suffix == ".gz":
        cg = ANNCandidateGenerator.from_gzip(path)
    else:
        cg = ANNCandidateGenerator.from_disk(path)
    return lambda: cg


# Should an abstract candidate generator be passing mentioncandidates that contain reference
# to the terms they matched on, and their respective similarities? If an candidate generator is really supposed to be
# returning conceptual candidates, returning this information seems like an implementation detail.
# Mention Candidates no longer return term matches, only concept id and its similarity to the span.
