from annotation import Annotation

from typing import List

from spacy.tokens import Span


class ContextWindow:
    """
    ContextWindow

    A ContextWindow holds a span of text no larger than 128 characters which is used for the context in getting embeddings.

    ------
    PARAMS:

    start_idx: int
        The index of the first character of the context window within the document
    end_idx: int
        The index of the character following the context window within the document
    annotations: List[Annotation]
        The list of annotations that lie within the context window. If this is left blank, then the annotations will be made from SpaCy's built-in ents property.
    embedding_cache: Dict[int, List[float]]
        A cache of embeddings for tokens in the window indexed by document-level index
    """

    def __init__(self, span: Span, annotations: List[Annotation] = []):
        self.span = span
        self.annotations = annotations
        for entity_span in self.span.ents:
            self.annotations.append(Annotation(entity_span))

        self.embedding_cache = {}

    def are_annotations_valid(self):
        """
        validate_annotations

        Ensures that each of the annotations in the context window is itself valid, and that none of the context windows overlap

        @seealso: annotation.is_valid()
        """
        for annotation in self.annotations:
            if not annotation.is_valid():
                return False

        # Sort annotations in order of starting index
        self.annotations = sorted(
            self.annotations, key=lambda annotation: annotation.span.start
        )

        prior_start_idx = self.span.start

        for annotation in self.annotations:
            if annotation.span.start < prior_start_idx:
                return False
            prior_start_idx = annotation.span.end

        return True

    def clean_unlinked_annotations(self):
        """
        clean_unlinked_annotations

        Removes annotations that have no viable candidate links
        """
        new_annotations = []
        for annotation in self.annotations:
            if len(annotation.candidates) != 0:
                new_annotations.append(annotation)

        self.annotations = new_annotations

    def get_structure_string(self):
        """
        get_structure_string

        Returns a formatted list of all the annotations in this window and their best candidates.
        """
        representation = " - " + self.span.text + "\n"
        for annotation in self.annotations:
            representation += (
                "\t - "
                + str(annotation)
                + " "
                + str(annotation.get_best_candidate())
                + "\n"
            )
        return representation

    def get_annotated_string(self, document: "Document", clear_newlines: bool = True):
        """
        get_annotated_string

        Gets a readable string representation of the window, with annotations displayed

        --------
        PARAMS:

        document: Document
            The document in which the context window lies
        clear_newlines: bool
            If true, all newlines in the text will be replaced with spaces. If false, only newlines in mentions will be replaced with spaces.
        """
        if len(self.annotations) == 0:
            if clear_newlines:
                return self.span.text.replace("\n", " ")
            return self.span.text

        representation = ""

        # Sort annotations in order of starting index
        annotations_to_add = sorted(
            self.annotations, key=lambda annotation: annotation.span.start
        )

        next_annotation_index = annotations_to_add[0].span.start
        prev_annotation_index = self.span.start

        for index in range(self.span.start, self.span.end):
            if next_annotation_index is not None and index == next_annotation_index:
                unannotated_section = document.spacy_doc[
                    prev_annotation_index:index
                ].text
                annotated_section = annotations_to_add[0].get_annotated_string()

                if len(unannotated_section) != 0:
                    representation += (
                        unannotated_section + " " + annotated_section + " "
                    )
                else:
                    representation += annotated_section + " "

                prev_annotation_index = annotations_to_add[0].span.end
                annotations_to_add.pop(0)

                if len(annotations_to_add) > 0:
                    next_annotation_index = annotations_to_add[0].span.start
                else:
                    next_annotation_index = None

        unannotated_section = document.spacy_doc[
            prev_annotation_index : self.span.end
        ].text

        representation += unannotated_section

        if clear_newlines:
            return representation.replace("\n", " ")

        return representation

    def remove_nested_annotations(self):
        """
        remove_nested_annotations

        Removes all annotations which intersect with larger annotations. If two annotations are of equal size, the earlier one is kept.
        """

        # Sort annotations by size (number of tokens)
        sorted_annotations = sorted(
            self.annotations,
            key=lambda annotation: annotation.span.end - annotation.span.start,
            reverse=True,
        )

        cleaned_annotations = []
        for candidate_annotation in sorted_annotations:
            nested = False
            # print("testing '{}'".format(candidate_annotation))
            for accepted_annotation in cleaned_annotations:
                if (
                    accepted_annotation.span.start
                    <= candidate_annotation.span.start
                    < accepted_annotation.span.end
                    or accepted_annotation.span.start
                    <= candidate_annotation.span.end
                    < accepted_annotation.span.end
                ):
                    nested = True
                    # print("intersects with '{}'".format(accepted_annotation))

            if not nested:
                cleaned_annotations.append(candidate_annotation)

        self.annotations = cleaned_annotations

    def cache_embedding(self, token_index: int, embedding: List[float]):
        """ "
        cache_embedding

        stores an embedding for a token in this context window's cache

        ---------
        PARAMS

        token_index: int
            The document-level index of the token associated with the embedding
        embedding: List[float]
            The embedding to cache
        """
        self.embedding_cache[token_index] = embedding

    def get_embedding_at(self, index: int):
        """
        get_embedding_at

        Returns a cached embedding at an index

        --------
        PARAMS:

        index: int
            The document-level index of the embedding to retrieve
        """
        return self.embedding_cache[index]
