from typing import List, Callable
from entity_type import EntityType


class Distancer:
    """
    Distancer

    Distancer is a class of static methods of calculating distance between two embeddings. Of these, only distance_between should actually be used, but the implementation can be swapped with any of the private distancing methods.

    """

    @staticmethod
    def distance_between(a: List[float], b: List[float]) -> float:
        """
        distance_between

        Returns the distance between the embeddings a and b by using one of the pre-implemented distance functions
        """
        return Distancer._euclidean_distance(a, b)

    @staticmethod
    def similarity_score(a: List[float], b: List[float]) -> float:
        return Distancer.dot_product(a, b)

    @staticmethod
    def get_all_labels_over_score(
        target: List[float], candidates: List[List[float]], threshhold
    ) -> float:
        """
        get_all_labels_over_score

        From a list of candidate embeddings, returns those whose similarity score to a target embedding exceeds the threshhold
        """

        condition = (
            lambda candidate: Distancer.similarity_score(target, candidate.embedding)
            > threshhold
        )

        close_candidates = filter(condition, candidates)

        return close_candidates

    @staticmethod
    def dot_product(a: List[float], b: List[float]) -> float:
        """
        dot_product

        Returns the dot product of two embeddings
        """
        sum = 0
        for a_term, b_term in zip(a, b):
            sum += a_term * b_term
        return sum

    @staticmethod
    def get_all_labels_within(
        target: List[float], candidates: List[EntityType], threshhold: float
    ):
        """
        get_all_labels_within

        Returns all the candidate entity_types from a list of entity types whose distance is under a given threshhold
        """
        condition = (
            lambda candidate: Distancer.distance_between(target, candidate.embedding)
            < threshhold
        )

        close_candidates = filter(condition, candidates)

        return close_candidates

    @staticmethod
    def get_closest(target: List[float], candidates: List[List[float]]):
        """
        get_closest

        Gets the closest embedding to a target embedding out of a list of candidates, measuring by using the default distance function
        """
        return Distancer._get_closest_using(
            Distancer.distance_between, target, candidates
        )

    @staticmethod
    def _get_closest_using(
        distanceFunction: Callable, target: List[float], candidates: List[List[float]]
    ):
        """
        _get_closest_using

        Gets the closest embedding to a target embedding out of a list of candidates, measuring by using a given distance function
        """
        closest_candidate = None
        closest_candidate_distance = None
        for candidate in candidates:
            candidate_distance = distanceFunction(candidate, target)
            if (closest_candidate is None) or (
                closest_candidate_distance < candidate_distance
            ):
                closest_candidate_distance = candidate_distance
                closest_candidate = candidate
        return closest_candidate

    @staticmethod
    def _chebyshev_distance(a: List[float], b: List[float]) -> float:
        """
        _chebyshev_distance

        Gets the chebyshev distance between embeddings with a given degree
        """

        if len(a) != len(b):
            raise ValueError("cannot compare vectors of unequal dimension")

        distance = 0.0

        for a_term, b_term in zip(a, b):
            distance = max(distance, abs(a_term - b_term))

        return distance

    @staticmethod
    def _euclidean_distance(a: List[float], b: List[float]) -> float:
        """
        _euclidean_distance

        Gets the euclidean distance between embeddings with a given degree
        """
        return Distancer._minkowski_distance(a, b, 2)

    @staticmethod
    def _manhattan_distance(a: List[float], b: List[float]) -> float:
        """
        _manhattan_distance

        Gets the manhattan distance between embeddings
        """
        return Distancer._minkowski_distance(a, b, 1)

    @staticmethod
    def _minkowski_distance(a: List[float], b: List[float], degree: int = 2):
        """
        _minkowski_distance

        Gets the minkowski distance between embeddings with a given degree
        """
        if len(a) != len(b):
            raise ValueError("cannot compare vectors of unequal dimension")

        distance_sum = 0.0

        for a_term, b_term in zip(a, b):
            distance_sum += abs(a_term - b_term) ** degree

        distance = distance_sum ** (1.0 / degree)
        return distance

    @staticmethod
    def _cross_encoder_distance(a, b):
        """
        _cross_encoder_distance

        Gets the distance between two embeddings by using a cross encoder
        """
        raise NotImplementedError("_cross_encoder_distance is not implmented")
