from context_window import ContextWindow
from windower import Windower

from typing import List

from spacy.tokens import Doc


class Document:
    """
    Document

    The Document class defines a document, including its context windows

    --------
    PARAMS:

    spacy_doc: Doc
        The SpaCy Doc that represents the underlying text
    windows: str
        The list of context windows that make up the document. If left empty, then the Windower class is used to create windows.
    sentences: List[Span]
        The list of sentences that compose the document, stored as a list rather than a potentially inconvenient generator
    """

    def __init__(self, spacy_doc: Doc, windows: List[ContextWindow] = []):
        self.spacy_doc = spacy_doc
        self.sentences = list(spacy_doc.sents)
        if windows == []:
            windower = Windower(self)
            self.windows = windower.get_windows()
        else:
            self.windows = windows

    def get_token_count(self):
        """
        get_token_count

        Returns the number of tokens in the document
        """
        return len(self.spacy_doc)

    def clean_unlinked_annotations(self):
        """
        clean_unlinked_annotations

        Removes annotations that have no viable candidate links
        """
        for window in self.windows:
            window.clean_unlinked_annotations()

    def get_structure_string(self):
        """
        get_structure_string

        Returns a formatted list of all the windows in this document, all their annotations and their best candidates.
        """
        representation = ""
        for window in self.windows:
            representation += window.get_structure_string()
        return representation

    def get_annotated_string(self, clear_newlines: bool = True):
        """
        get_annotated_string

        Gets a readable string representation of the document, with annotations displayed
        """

        representation = ""
        for window in self.windows:
            representation += window.get_annotated_string(self, clear_newlines) + " "
        return representation

    def remove_nested_annotations(self):
        """
        remove_nested_annotations

        From each context window, removes annotations which intersect with larger annotations.
        """
        for window in self.windows:
            window.remove_nested_annotations()
