from typing import Iterable, Callable, Any, List
import math

from spacy.tokens import Span
from distancer import Distancer
from annotation import Annotation
from entity_type import EntityType
from document import Document
from context_window import ContextWindow
from candidate import Candidate
from embedding_utils import EmbeddingUtils

# from spacy.vocab import Vocab
# from spacy.training import Example
# from spacy.pipeline import TrainablePipe


class EntityLinker:
    """
    EntityLinker

    The entity linker is a trainable pipe to help assign links to entities
    """

    """
    def __init__(self, vocab: Vocab) -> None:
        super().__init__(vocab)
        raise NotImplementedError("EntityLinker is not implemented yet")

    def update(self, examples: Iterable[Example], **kwargs: Any):
        raise NotImplementedError("update is not implemented yet")

    def predict(self, examples: Iterable[Example]) -> Iterable[Example]:
        raise NotImplementedError("predict is not implemented yet")

    def initialize(self, get_examples: Callable[[], Iterable[Example]]):
        raise NotImplementedError("initialize is not implemented yet")

    def to_disk(self, path: pathlib.Path, exclude=tuple()):
        raise NotImplementedError("to_disk is not implemented yet")

    def from_disk(self, path: pathlib.Path, exclude=tuple()):
        raise NotImplementedError("from_disk is not implemented yet")
    """

    def __init__(self, dual_encoder: "DualEncoder", entity_labels: List[EntityType]):
        self.dual_encoder = dual_encoder

        self.entity_labels = entity_labels

        for entity_label in entity_labels:
            entity_label.embedding = dual_encoder.encode(entity_label.name)

    def link_entities_in_document(
        self, doc: Document, verbose: bool = False, show_progress_bar: bool = False
    ):
        """
        link_entities_in_document

        Recognizes the entity mentions in a document and assigns each annotation a list of possible links.

        -------
        PARAMS:

        doc: Document
            The document to assign context windows to
        verbose: bool
            If true, prints the best entity that was linked to each enitity mention
        show_progress_bar: bool
            If true, prints out a progress bar of the linking process. Cannot be enabled while verbose is true.
        """

        if verbose:
            show_progress_bar = False

        self.initialize_embeddings(doc, show_progress_bar)

        # if show_progress_bar:
        # print("|" + " " * len(doc.windows) + "|")
        # print("|", end="")

        for window in doc.windows:
            window.annotations = self.get_annotations_for_window(window, doc)

            if verbose:
                print("window is: '{}'".format(window.span.text.replace("\n", " ")))
            for annotation in window.annotations:
                annotation.candidates = self.get_candidates_for_annotation(
                    annotation, window
                )
                if verbose:
                    print(
                        "{} linked to {}".format(
                            annotation, annotation.get_best_candidate()
                        )
                    )

            window.clean_unlinked_annotations()
            window.remove_nested_annotations()
            # if show_progress_bar:
            # print("#", end="")
        # if show_progress_bar:
        # print("|")

    def get_annotations_for_window(self, window: ContextWindow, doc: Document):
        """
        get_annotations_for_window

        Finds enity mention annotations in a context window.

        ------------
        PARAMS:

        window: ContextWindow
            The window that annotations are assigned to
        """
        PROBABILITY_THRESH = 0.3  # 0.5 - 0.6 in paper
        annotations = []
        for length in range(1, 4):
            for i in range(window.span.start, window.span.end - length):
                annotation_span = doc.spacy_doc[i : i + length]
                probability = self.get_candidate_mention_span_probability(
                    annotation_span, window, doc
                )
                if probability > PROBABILITY_THRESH:
                    annotations.append(Annotation(annotation_span))

        return annotations

        """
        # Right now, just labels every token with len > 1 as an entity mention
        annotations = []
        for i in range(window.span.start, window.span.end):
            annotation_span = doc.spacy_doc[i : i + 1]
            if len(annotation_span.text) < 2:
                continue

            annotation = Annotation(annotation_span)
            annotations.append(annotation)
        return annotations
        """

    def get_candidates_for_annotation(
        self, annotation: Annotation, context: ContextWindow
    ):
        """
        get_candidates_for_annotation

        Finds a list of possible entity links for an entity mention

        --------
        PARAMS:

        annotation: Annotation
            The entity mention annotation to assign links to
        context: ContextWindow
            The context in which the annotation lies.
        """
        target_embedding = self.dual_encoder.encode(annotation.span.text, context)

        candidate_entities = Distancer.get_all_labels_over_score(
            target_embedding, self.entity_labels, 0.4
        )

        candidates = list(
            map(
                lambda candidate_entity: Candidate(
                    candidate_entity,
                    Distancer.similarity_score(
                        candidate_entity.embedding, target_embedding
                    ),
                ),
                candidate_entities,
            )
        )

        return candidates

    def initialize_embeddings(self, doc: Document, show_progress: bool = False):
        """
        initialize_embeddings

        Finds all the embeddings of all tokens in the document and stores them in the windows' respective caches.

        --------
        PARAMS:

        doc: Document
            The document to cache embeddings of
        """
        completed_count = 0
        for window in doc.windows:
            for i in range(window.span.start, window.span.end):
                embedding = self.dual_encoder.encode(doc.spacy_doc[i].text, window)
                window.cache_embedding(i, embedding)
                if show_progress:
                    completed_count += 1
                    print("{}%".format(100 * completed_count / len(doc.spacy_doc)))

    def get_candidate_mention_span_probability(
        self, mention_span: Span, context: ContextWindow, doc: Document
    ):
        """
        get_candidate_mention_span_probability

        Gets the probability of an n-gram representing an entity mention.
        Uses' Bhowmik's formula: P(m|(i,j)) = \sigma(w^T_sh_i +w^T_eh_j+\sum^{j}_{q=i}{(w_m^Th_q)})

        --------
        PARAMS:

        mention_span: Span
            The SpaCy span that represents the potential entity mention
        context: ContextWindow
            The context window in which the mention span lies
        doc: Document
            The document in which the mention span lies
        """
        EMBEDDING_LENGTH = 384
        w_mid = EmbeddingUtils.W_MID  # [1] * EMBEDDING_LENGTH
        w_start = [1] * EMBEDDING_LENGTH
        w_end = EmbeddingUtils.W_MID  # [1] * EMBEDDING_LENGTH

        mid_sum = 0
        for q in range(mention_span.start, mention_span.end):
            mid_sum += Distancer.dot_product(w_mid, context.get_embedding_at(q))

        start_sum = Distancer.dot_product(
            w_start, context.get_embedding_at(mention_span.start)
        )
        end_sum = Distancer.dot_product(
            w_end, context.get_embedding_at(mention_span.end)
        )

        total_sum = start_sum + end_sum + mid_sum

        sigmoid_result = 1.0 / (1.0 + math.exp(-total_sum))

        return sigmoid_result
