from typing import List, Any

# from document import Document
from context_window import ContextWindow


class Windower:
    """
    Windower

    The windower extracts context windows from a document. These windows may be based on sentences and cannot exceed 128 chars.

    ------
    PARAMS:

    document: Document
        The document to extract windows out of

    start_idx: int
        The index to start windowing at. Always begins at 0, increments as windows are extracted
    """

    def __init__(self, document: "Document"):
        self.document = document
        self.start_idx = 0

    def get_windows(self) -> List[ContextWindow]:
        """
        get_windows

        Gets a list of all the windows in the document
        """
        windows = []

        while not self.is_complete():
            windows.append(self.get_window())
        return windows

    def get_window(self):
        """
        get_window

        Gets the next context window in the document. Calls one of another implementations of window extraction.
        """

        # return self._get_token_window_under_char_count(128)
        # return self._get_sentence_window_under_char_count(128)
        return self._get_single_sentence_window()

    def is_complete(self) -> bool:
        """
        is_complete

        Returns whether the windower has extracted all possible windows in the document.
        """
        # return self.document.get_token_count() <= self.start_idx # for token-based indexing
        return len(self.document.sentences) <= self.start_idx

    def _get_token_window_under_char_count(self, count: int):
        """
        _get_token_window_under_char_count

        Gets the longest window composed of whole tokens whose text representation is less than a given char count

        count: int
            The maximum length, in characters, of the returned window.
        """
        window_end_index = self.start_idx
        while (
            len(self.document.spacy_doc[self.start_idx : window_end_index + 1].text)
            < count
            and not window_end_index > self.document.get_token_count()
        ):
            window_end_index += 1

        window_span = self.document.spacy_doc[self.start_idx : window_end_index]
        self.start_idx = window_end_index

        return ContextWindow(window_span)

    def _get_single_sentence_window(self):
        """
        _get_single_sentence_window

        Get a window whose length is a single sentence.
        """
        sentence = self.document.sentences[self.start_idx]
        window_span = Windower._get_range_span(self.document, sentence, sentence)
        self.start_idx += 1
        return ContextWindow(window_span)

    def _get_sentence_window_under_char_count(self, count: int):
        """
        _get_sentence_window_under_char_count

        Gets the longest window composed of whole sentences whose text representation is less than a given char count

        count: int
           The maximum length, in characters, of the returned window.
        """
        window_end_index = self.start_idx
        while Windower._char_length_of_spans(
            self.document.sentences[self.start_idx : window_end_index + 1]
        ) < count and not window_end_index > len(self.document.sentences):
            print(window_end_index)
            window_end_index += 1

        window_end_index = min(window_end_index, len(self.document.sentences) - 1)

        window_span = Windower._get_range_span(
            self.document,
            self.document.sentences[self.start_idx],
            self.document.sentences[window_end_index],
        )
        print(window_span.text)
        self.start_idx = window_end_index
        return ContextWindow(window_span)

    @staticmethod
    def _char_length_of_spans(spans: List["Span"]):
        """
        _char_length_of_spans

        A static method that gets the combined length, in characters, of a list of spans
        """
        char_length = 0
        for span in spans:
            char_length += len(span.text)
        return char_length

    @staticmethod
    def _get_range_span(document: "Document", startSpan: "Span", endSpan: "Span"):
        """
        _get_range_span

        Returns a new span which includes two spans and all the space in between

        -----------------
        PARAMS:

        document: Document
            The document in which the span lives
        startSpan: Span
            The first of the two spans to bridge
        endSpan: Span
            The last of the two spans to bridge
        """
        return document.spacy_doc[startSpan.start : endSpan.end]
