import itertools
from typing import Callable, Dict, Iterable, List, Literal, Optional, Tuple, cast

import numpy
import spacy
from spacy.errors import Errors
from spacy.language import Language
from spacy.ml import extract_spans
from spacy.pipeline import TrainablePipe
from spacy.tokens import Doc, Span, SpanGroup
from spacy.training import Example
from spacy.vocab import Vocab
from thinc.api import (
    Model,
    Optimizer,
    Ragged,
    Softmax,
    chain,
    list2ragged,
    set_dropout_rate,
    with_getitem,
)
from thinc.types import Floats2d, Ints2d


class ContainerHandler:
    def _get_containers(self, doc: Doc) -> Iterable[Span]:
        raise NotImplementedError

    def add_containers(self, doc: Doc, span_group: SpanGroup) -> None:
        raise NotImplementedError


class EntHandler(ContainerHandler):
    def get_containers(self, doc: Doc) -> Iterable[Span]:
        return doc.ents

    def add_containers(self, doc: Doc, span_group: SpanGroup) -> None:
        doc.set_ents(list(span_group))


class SpanHandler(ContainerHandler):
    def __init__(self, span_key: str = "sc"):
        self._span_key = span_key

    def get_containers(self, doc: Doc) -> Iterable[Span]:
        return doc.spans[self._span_key]

    def add_containers(self, doc: Doc, span_group: SpanGroup) -> None:
        doc.spans[self._span_key] = span_group


class PreExtractedRecognizer(TrainablePipe):
    """This class is a base class and not instantiated directly. Trainable
    pipeline components like the EntityRecognizer or TextCategorizer inherit
    from it and it defines the interface that components should follow to
    function as trainable components in a spaCy pipeline.
    DOCS: https://spacy.io/api/pipe
    """

    def __init__(
        self,
        vocab: Vocab,
        model: Model,
        name: str,
        ent_or_span: Literal["ENT", "SPAN"],
        span_key: Optional[str] = None,
        **cfg,
    ):
        """Initialize a pipeline component.
        vocab (Vocab): The shared vocabulary.
        model (thinc.api.Model): The Thinc Model powering the pipeline component.
        name (str): The component instance name.
        **cfg: Additional settings and config parameters.
        DOCS: https://spacy.io/api/pipe#init
        """
        self.vocab = vocab
        self.model = model
        self.name = name
        self.cfg = dict(cfg)
        if self._ent_or_span == "ENT":
            self._container_handler: ContainerHandler = EntHandler()
        elif self._ent_or_span == "SPAN":
            assert span_key
            self._container_handler = SpanHandler(span_key)

    def set_annotations(self, docs: Iterable[Doc], indices_scores) -> None:
        labels = self.labels
        indices, scores = indices_scores
        offset = 0
        for i, doc in enumerate(docs):
            indices_i = indices[i].dataXd
            span_group = self._make_span_group(
                doc, indices_i, scores[offset : offset + indices.lengths[i]], labels  # type: ignore[arg-type]
            )
            offset += indices.lengths[i]
            self._container_handler.add_containers(doc, span_group)

    def add_label(self, label: str) -> int:
        if not isinstance(label, str):
            raise ValueError(Errors.E187)
        if label in self.labels:
            return 0
        self._allow_extra_label()
        self.cfg["labels"].append(label)  # type: ignore[attr-defined]
        if self.model and "resize_output" in self.model.attrs:
            self.model = self.model.attrs["resize_output"](self.model, len(self.labels))
        self.vocab.strings.add(label)
        return 1

    def predict(self, docs: Iterable[Doc]):
        indices = self._get_ragged_data(docs)
        if indices.lengths.sum() == 0:
            scores = self.model.ops.alloc2f(0, 0)
        else:
            scores = self.model.predict((docs, indices))  # type: ignore
        return indices, scores

    def _get_ragged_data(self, docs: Iterable[Doc]) -> Ragged:
        ops = self.model.ops
        containers = itertools.chain.from_iterable(
            (self._get_containers(doc) for doc in docs)
        )
        indices = [(container.start, container.end) for container in containers]
        lengths = ops.asarray1i([len(container) for container in containers])
        if len(indices) > 0:
            output = Ragged(ops.xp.vstack(indices), lengths)
        else:
            output = Ragged(ops.xp.zeros((0, 0), dtype="i"), lengths)
        assert output.dataXd.ndim == 2
        return output

    def update(
        self,
        examples: Iterable[Example],
        *,
        drop: float = 0.0,
        sgd: Optional[Optimizer] = None,
        losses: Optional[Dict[str, float]] = None,
    ) -> Dict[str, float]:
        if losses is None:
            losses = {}
        losses.setdefault(self.name, 0.0)
        # validate_examples(examples, "SpanCategorizer.update")
        self._validate_categories(examples)
        if not any(len(eg.predicted) if eg.predicted else 0 for eg in examples):
            # Handle cases where there are no tokens in any docs.
            return losses
        docs = [eg.predicted for eg in examples]
        spans = self._get_ragged_data(docs)
        if spans.lengths.sum() == 0:
            return losses
        set_dropout_rate(self.model, drop)
        scores, backprop_scores = self.model.begin_update((docs, spans))
        loss, d_scores = self.get_loss(examples, (spans, scores))
        backprop_scores(d_scores)  # type: ignore
        if sgd is not None:
            self.finish_update(sgd)
        losses[self.name] += loss
        return losses

    def get_loss(
        self, examples: Iterable[Example], span_scores: Tuple[Ragged, Floats2d]
    ) -> Tuple[float, float]:
        spans, scores = span_scores
        spans = Ragged(
            self.model.ops.to_numpy(spans.data), self.model.ops.to_numpy(spans.lengths)
        )
        label_map = {label: i for i, label in enumerate(self.labels)}
        target = numpy.zeros(scores.shape, dtype=scores.dtype)
        offset = 0
        for i, eg in enumerate(examples):
            # Map (start, end) offset of spans to the row in the d_scores array,
            # so that we can adjust the gradient for predictions that were
            # in the gold standard.
            spans_index = {}
            spans_i = spans[i].dataXd
            for j in range(spans.lengths[i]):
                start = int(spans_i[j, 0])  # type: ignore
                end = int(spans_i[j, 1])  # type: ignore
                spans_index[(start, end)] = offset + j
            for gold_span in self._get_aligned_spans(eg):
                key = (gold_span.start, gold_span.end)
                if key in spans_index:
                    row = spans_index[key]
                    k = label_map[gold_span.label_]
                    target[row, k] = 1.0
            # The target is a flat array for all docs. Track the position
            # we're at within the flat array.
            offset += spans.lengths[i]
        target = self.model.ops.asarray(target, dtype="f")  # type: ignore
        # The target will have the values 0 (for untrue predictions) or 1
        # (for true predictions).
        # The scores should be in the range [0, 1].
        # If the prediction is 0.9 and it's true, the gradient
        # will be -0.1 (0.9 - 1.0).
        # If the prediction is 0.9 and it's false, the gradient will be
        # 0.9 (0.9 - 0.0)
        d_scores = scores - target
        loss = float((d_scores**2).sum())
        return loss, d_scores

    @property
    def labels(self) -> Tuple[str]:
        """RETURNS (Tuple[str]): The labels currently added to the component.
        DOCS: https://spacy.io/api/textcategorizer#labels
        """
        return tuple(self.cfg["labels"])  # type: ignore[arg-type, return-value]

    def initialize(
        self,
        get_examples: Callable[[], Iterable[Example]],
        *,
        labels: Optional[List[str]] = None,
        nlp: Language,
    ):
        subbatch: List[Example] = []
        if labels is not None:
            for label in labels:
                self.add_label(label)
        for eg in get_examples():
            if labels is None:
                for span in eg.reference.spans.get(self.key, []):
                    self.add_label(span.label_)
            if len(subbatch) < 10:
                subbatch.append(eg)
        self._require_labels()
        if subbatch:
            docs = [eg.x for eg in subbatch]
            spans = self._get_ragged_data(docs)
            Y = self.model.ops.alloc2f(spans.dataXd.shape[0], len(self.labels))
            self.model.initialize(X=(docs, spans), Y=Y)
        else:
            self.model.initialize()

    @property
    def is_trainable(self) -> bool:
        return True

    def _make_span_group(
        self, doc: Doc, indices: Ints2d, scores: Floats2d, labels: List[str]
    ) -> SpanGroup:
        spans = SpanGroup(doc, name=self.key)
        winning_label_indices = (scores * -1).argmax()  # type: ignore
        for i in range(winning_label_indices):
            start = indices[i, 0]
            end = indices[i, 1]
            label = labels[winning_label_indices]
            spans.append(Span(doc, start, end, label))
        return spans


@spacy.registry.architectures("pre_extracted_recognizer")
def build_pre_extracted_recognizer(
    tok2vec: Model[List[Doc], List[Floats2d]],
    reducer: Model[Ragged, Floats2d],
    nO=None,
    nI=None,
) -> Model[Tuple[List[Doc], Ragged], Floats2d]:
    model = chain(
        cast(
            Model[Tuple[List[Doc], Ragged], Tuple[Ragged, Ragged]],
            with_getitem(
                0, chain(tok2vec, cast(Model[List[Floats2d], Ragged], list2ragged()))
            ),
        ),
        extract_spans(),
        reducer,
        Softmax(nO=nO, nI=nI),
    )
    model.set_ref("tok2vec", tok2vec)
    model.set_ref("reducer", reducer)
    return model
