from collections import defaultdict
import pathlib
from crtl.lkb.lkb import ConceptRelation, LexicalKnowledgeBase
from crtl.components.ann_vector_kb import ANNCandidateGenerator


class APIController:
    def __init__(
        self, lkb: LexicalKnowledgeBase, no_context_entity_linker: ANNCandidateGenerator
    ):
        self._lkb = lkb
        self._ncel = no_context_entity_linker
        self._terms = [term for term in list(lkb.term_to_cuis.keys()) if term]
        self._terms.sort()
        self._term_index = {term: idx for idx, term in enumerate(self._terms)}
        self._inverse_concept_rels: dict[str, list[ConceptRelation]] = defaultdict(list)
        for concept in self._lkb.cui_to_concept.values():
            concept_relations = concept.related_concepts
            for relation in concept_relations:
                self._inverse_concept_rels[relation.concept_id].append(
                    ConceptRelation(relation.relation, concept.concept_id)
                )

    def get_concept_info(self, concept_id) -> dict[str]:
        concept = self._lkb.cui_to_concept[concept_id]
        terms = [
            {"term_id": self._term_index[term], "term_string": term}
            for term in concept.terms
        ]
        # Bring "isa" to the front

        inward_edges = self._inverse_concept_rels[concept_id]
        inverse_isa_edges = filter(
            lambda rel: rel.relation == "inverse_isa", inward_edges
        )
        non_inverse_isa_edges = filter(
            lambda rel: rel.relation != "inverse_isa", inward_edges
        )
        inward_related_concepts: dict[str, str] = [
            {
                "relation_string": rel.relation,
                "concept_id": rel.concept_id,
                "concept_name": self._lkb.cui_to_concept[rel.concept_id].canonical_name,
            }
            for rel in list(inverse_isa_edges) + list(non_inverse_isa_edges)
        ]
        # Bring "inverse_isa" to the front
        outward_edges = concept.related_concepts
        isa_edges = filter(lambda rel: rel.relation == "isa", outward_edges)
        non_isa_edges = filter(lambda rel: rel.relation != "isa", outward_edges)
        outward_related_concepts: dict[str, str] = [
            {
                "relation_string": rel.relation,
                "concept_id": rel.concept_id,
                "concept_name": self._lkb.cui_to_concept[rel.concept_id].canonical_name,
            }
            for rel in list(isa_edges) + list(non_isa_edges)
        ]
        return {
            "id": concept.concept_id,
            "concept_name": concept.canonical_name,
            "semantic_type": "placeholder",
            "definition": concept.definition,
            "terms": terms,
            "inward_related_concepts": inward_related_concepts,
            "outward_related_concepts": outward_related_concepts,
        }

    def get_term_info(self, term_id):
        term = self._terms[int(term_id)]
        concepts = [
            self._lkb.cui_to_concept[cui] for cui in self._lkb.term_to_cuis[term]
        ]
        return {
            "term_id": term_id,
            "term_string": term,
            "concepts": [
                {
                    "concept_id": concept.concept_id,
                    "concept_name": concept.canonical_name,
                }
                for concept in concepts
            ],
        }

    def search_for_concept(self, search_term):
        concepts = [
            self._lkb.cui_to_concept[mention_candidate.concept_id]
            for mention_candidate in self._ncel([search_term], k=10)[0]
        ]
        return [
            {"concept_id": concept.concept_id, "concept_name": concept.canonical_name}
            for concept in concepts
        ]


def create_trendnet_api_controller():
    # trendnet_linker_path = LINKER_PATHS[trendnet_name]
    trendnet = LexicalKnowledgeBase.from_disk(
        "/home/ubuntu/CRTL/src/crtl/lkb/trendnet_2.jsonl"
    )
    ann_cg = ANNCandidateGenerator.from_disk(
        pathlib.Path("/home/ubuntu/CRTL/src/crtl/lkb/trendnet_2_ann_cg")
    )

    return APIController(trendnet, ann_cg)


if __name__ == "__main__":
    api_controller = create_trendnet_api_controller()

# print me
