"""
Convert a umls release to a jsonl file of concepts. Originally coming from https://github.com/allenai/scispacy/blob/main/scripts/export_umls_json.py
"""
import json
import argparse
import pathlib
from typing import List, Dict, Optional


def create_umls_json_from_meta(
    meta_path: pathlib.Path, output_path: pathlib.Path, source: Optional[str] = None
):
    concept_details: Dict = {}  # dictionary of concept_id -> {
    #                 'concept_id': str,
    #                 'canonical_name': str
    #                 'terms': List[str]
    #                 'types': List[str]
    #                 'definition': str
    # }

    print("Reading concepts ... ")
    read_umls_concepts(meta_path, concept_details, source)

    print("Reading types ... ")
    read_umls_types(meta_path, concept_details)

    print("Reading definitions ... ")
    read_umls_definitions(meta_path, concept_details)

    print("Reading concept relations ... ")
    read_umls_relations(meta_path, concept_details, source)

    without_canonical_name_count = 0
    without_terms_count = 0
    with_one_term_count = 0
    with_more_than_one_term_count = 0
    without_type_count = 0
    with_one_type_count = 0
    with_more_than_one_type_count = 0
    without_definition_count = 0
    with_definition_pref_source_count = 0
    with_definition_other_sources_count = 0
    for concept in concept_details.values():
        without_canonical_name_count += 1 if "canonical_name" not in concept else 0
        without_terms_count += 1 if len(concept["terms"]) == 0 else 0
        with_one_term_count += 1 if len(concept["terms"]) == 1 else 0
        with_more_than_one_term_count += 1 if len(concept["terms"]) > 1 else 0
        without_type_count += 1 if len(concept["types"]) == 0 else 0
        with_one_type_count += 1 if len(concept["types"]) == 1 else 0
        with_more_than_one_type_count += 1 if len(concept["types"]) >= 1 else 0
        without_definition_count += 1 if "definition" not in concept else 0
        with_definition_pref_source_count += (
            1 if concept.get("is_from_preferred_source") == "Y" else 0
        )
        with_definition_other_sources_count += (
            1 if concept.get("is_from_preferred_source") == "N" else 0
        )

    print(f"Number of concepts: {len(concept_details)}")
    print(
        f"Number of concepts without canonical name (one of the terms will be used instead): "
        f"{without_canonical_name_count}"
    )
    print(f"Number of concepts with no terms: {without_terms_count}")
    print(f"Number of concepts with 1 term: {with_one_term_count}")
    print(f"Number of concepts with > 1 term: {with_more_than_one_term_count}")
    print(f"Number of concepts with no type: {without_type_count}")
    print(f"Number of concepts with 1 type: {with_one_type_count}")
    print(f"Number of concepts with > 1 type: {with_more_than_one_type_count}")
    print(f"Number of concepts with no definition: {without_definition_count}")
    print(
        f"Number of concepts with definition from preferred sources: {with_definition_pref_source_count}"
    )
    print(
        f"Number of concepts with definition from other sources: {with_definition_other_sources_count}"
    )

    print("Deleting unused fields and choosing a canonical name from terms ... ")
    for concept in concept_details.values():
        # Some concepts have many duplicate terms. Here we remove them.
        concept["terms"] = list(set(concept["terms"]))

        # if a concept doesn't have a canonical name, use the first term instead
        if "canonical_name" not in concept:
            terms = concept["terms"]
            concept["canonical_name"] = terms[0]

        # deleting `is_from_preferred_source`
        if "is_from_preferred_source" in concept:
            del concept["is_from_preferred_source"]

    print("Exporting to the jsonl file {} ...".format(output_path))
    with open(output_path, "w") as fout:
        for value in concept_details.values():
            fout.write(json.dumps(value) + "\n")
    print("DONE.")


# preferred definition sources (from S2)
DEF_SOURCES_PREFERRED = {"NCI_BRIDG", "NCI_NCI-GLOSS", "NCI", "GO", "MSH", "NCI_FDA"}


def read_umls_file_headers(meta_path: pathlib.Path, filename: str) -> List[str]:
    """
    Read the file descriptor MRFILES.RRF from a UMLS release and get column headers (names)
    for the given file
    MRFILES.RRF file format: a pipe-separated values
    Useful columns:
        column 0: name of one of the files in the META directory
        column 2: column names of that file
    Args:
        meta_path: path to the META directory of an UMLS release
        filename: name of the file to get its column headers
    Returns:
        a list of column names
    """
    file_descriptors = f"{meta_path}/MRFILES.RRF"  # to get column names
    with open(file_descriptors, encoding="utf-8") as fin:
        for line in fin:
            splits = line.split("|")
            found_filename = splits[0]
            column_names = (splits[2] + ",").split(
                ","
            )  # ugly hack because all files end with an empty column
            if found_filename in filename:
                return column_names
    assert False, f"Couldn't find column names for file {filename}"
    return None


def read_umls_concepts(
    meta_path: pathlib.Path, concept_details: Dict, source: Optional[str] = None
):
    """
    Read the concepts file MRCONSO.RRF from a UMLS release and store it in
    concept_details dictionary. Each concept is represented with
    - concept_id
    - canonical_name
    - terms
    - types
    - definition
    This function fills the first three. If a canonical name is not found, it is left empty.
    MRFILES.RRF file format: a pipe-separated values
    Useful columns: CUI, LAT, SUPPRESS, STR, ISPREF, TS, STT
    Args:
        meta_path: path to the META directory of an UMLS release
        concept_details: a dictionary to be filled with concept informations
        source: An optional source identifier, used as a filter to extract only a
                specific source from UMLS.
    """
    concepts_filename = "MRCONSO.RRF"
    headers = read_umls_file_headers(meta_path, concepts_filename)
    with open(f"{meta_path}/{concepts_filename}", encoding="utf-8") as fin:
        for line in fin:
            splits = line.strip().split("|")
            assert len(headers) == len(splits), (headers, splits)
            concept = dict(zip(headers, splits))
            if concept["LAT"] != "ENG" or concept["SUPPRESS"] != "N":
                continue  # Keep English non-suppressed concepts only

            if source is not None:
                if concept["SAB"] != source:
                    continue

            concept_id = concept["CUI"]
            if concept_id not in concept_details:  # a new concept
                # add it to the dictionary with an empty list of terms and types
                concept_details[concept_id] = {
                    "concept_id": concept_id,
                    "terms": [],
                    "types": [],
                }

            concept_name = concept["STR"]
            # this condition is copied from S2. It checks if the concept name is canonical or not
            is_canonical = (
                concept["ISPREF"] == "Y"
                and concept["TS"] == "P"
                and concept["STT"] == "PF"
            )
            concept_details[concept_id]["terms"].append(
                concept_name
            )  # add it as an term
            if is_canonical and "canonical_name" not in concept_details[concept_id]:
                concept_details[concept_id][
                    "canonical_name"
                ] = concept_name  # set as canonical name
            concept_details[concept_id]["related_concepts"] = []


def read_umls_types(meta_path: pathlib.Path, concept_details: Dict):
    """
    Read the types file MRSTY.RRF from a UMLS release and store it in
    concept_details dictionary. This function adds the `types` field
    to the information of each concept
    MRSTY.RRF file format: a pipe-separated values
    Useful columns: CUI, TUI
    Args:
        meta_path: path to the META directory of an UMLS release
        concept_details: a dictionary to be filled with concept informations
    """
    types_filename = "MRSTY.RRF"
    headers = read_umls_file_headers(meta_path, types_filename)
    with open(f"{meta_path}/{types_filename}", encoding="utf-8") as fin:
        for line in fin:
            splits = line.strip().split("|")
            assert len(headers) == len(splits)
            concept_type = dict(zip(headers, splits))

            concept = concept_details.get(concept_type["CUI"])
            if (
                concept is not None
            ):  # a small number of types are for concepts that don't exist
                concept["types"].append(concept_type["TUI"])


def read_umls_definitions(meta_path: pathlib.Path, concept_details: Dict):
    """
    Read the types file MRDEF.RRF from a UMLS release and store it in
    concept_details dictionary. This function adds the `definition` field
    to the information of each concept
    MRDEF.RRF file format: a pipe-separated values
    Useful columns: CUI, SAB, SUPPRESS, DEF
    Args:
        meta_path: path to the META directory of an UMLS release
        concept_details: a dictionary to be filled with concept informations
    """
    definitions_filename = "MRDEF.RRF"
    headers = read_umls_file_headers(meta_path, definitions_filename)
    with open(f"{meta_path}/{definitions_filename}", encoding="utf-8") as fin:
        headers = read_umls_file_headers(meta_path, definitions_filename)
        for line in fin:
            splits = line.strip().split("|")
            assert len(headers) == len(splits)
            definition = dict(zip(headers, splits))

            if definition["SUPPRESS"] != "N":
                continue
            is_from_preferred_source = definition["SAB"] in DEF_SOURCES_PREFERRED
            concept = concept_details.get(definition["CUI"])
            if (
                concept is None
            ):  # a small number of definitions are for concepts that don't exist
                continue

            if (
                "definition" not in concept
                or is_from_preferred_source
                and concept["is_from_preferred_source"] == "N"
            ):
                concept["definition"] = definition["DEF"]
                concept["is_from_preferred_source"] = (
                    "Y" if is_from_preferred_source else "N"
                )


def read_umls_relations(
    meta_path: pathlib.Path, concept_details: Dict, source: Optional[str] = None
):
    """
    Read the types file MRREL.RRF from a UMLS release and store it in
    concept_details dictionary. This function adds the `related_concepts` field
    to the information of each concept
    MRREL.RRF file format: a pipe-separated values
    Useful columns: CUI, SAB, SUPPRESS, DEF
    Args:
        meta_path: path to the META directory of an UMLS release
        concept_details: a dictionary to be filled with concept informations
    """
    relations_filename = "MRREL.RRF"

    with open(f"{meta_path}/{relations_filename}", encoding="utf-8") as fin:
        headers = read_umls_file_headers(meta_path, relations_filename)
        for line in fin:
            splits = line.strip().split("|")
            assert len(headers) == len(splits)
            relation = dict(zip(headers, splits))

            if relation["SUPPRESS"] != "N":
                continue
            if source:
                if relation["SAB"] != source:
                    continue
            # need to do something more clever to make sure only related concepts from source vocabs are included.
            concept = concept_details.get(relation["CUI2"])
            if concept and relation["STYPE1"] == relation["STYPE2"] == "SCUI":
                concept["related_concepts"].append(
                    {"relation": relation["RELA"], "concept_id": relation["CUI1"]}
                )


if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "--meta_path", help="Path to the META directory of an UMLS release."
    )
    parser.add_argument("--output_path", help="Path to the output jsonl file")
    parser.add_argument(
        "--source",
        type=str,
        default=None,
        help="Whether to filter for a only a single UMLS source.",
    )
    args = parser.parse_args()
    create_umls_json_from_meta(args.meta_path, args.output_path, args.source)
