import argparse
import os
import pathlib
import requests
from zipfile import ZipFile
from tqdm import tqdm


def download_umls_metathesaurus(
    umls_version: str, output_directory: pathlib.Path, subset_name="metathesaurus-full"
):
    zipped_output_path = output_directory / (umls_version + ".zip")
    params = (
        (
            "url",
            f"https://download.nlm.nih.gov/umls/kss/{umls_version}/umls-{umls_version}-{subset_name}.zip",
        ),
        ("apiKey", os.environ["UMLS_API_KEY"]),
    )
    response = requests.get(
        "https://uts-ws.nlm.nih.gov/download", params=params, stream=True
    )

    # NB. Original query string below. It seems impossible to parse and
    # reproduce query strings 100% accurately so the one below is given
    # in case the reproduced version is not "correct".
    # response = requests.get('https://uts-ws.nlm.nih.gov/download?url=https://download.nlm.nih.gov/umls/kss/2022AA/umls-2022AA-full.zip&apiKey=YOUR_API_KEY')
    total_size_in_bytes = int(
        response.headers.get("content-length", 0)
    )  # should be ~2-3 GB, content-length is empty unfortunately so pbar won't have total data size
    block_size = 1024  # 1 Kibibyte
    with open(zipped_output_path, "wb") as f:
        with tqdm(total=total_size_in_bytes, unit="iB", unit_scale=True) as pbar:
            for data in response.iter_content(block_size):
                f.write(data)
                pbar.update(len(data))
    # unzip the response
    with ZipFile(zipped_output_path, "r") as zObject:
        zObject.extractall(output_directory)
    os.remove(zipped_output_path)


if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "--umls_version", help="Path to the META directory of an UMLS release."
    )
    parser.add_argument("--output_directory", help="Path to the output file")
    parser.add_argument(
        "--subset_name",
        help="Name of UMLS metathesaurus subset: defaults to metathesaurus-full",
    )
    args = parser.parse_args()
    download_umls_metathesaurus(args.umls_version, pathlib.Path(args.output_directory))
